@@ -2,7 +2,7 @@
 
 #ifndef __SCMH
 #define __SCMH
-/*	Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+/*	Copyright (C) 1995, 1996, 1998 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -48,17 +48,17 @@
  * These may be defined or undefined.
  */
 
-/* If the compile FLAG `CAUTIOUS' is #defined then the number of
+/* If the compile FLAG `SCM_CAUTIOUS' is #defined then the number of
  * arguments is always checked for application of closures.  If the
- * compile FLAG `RECKLESS' is #defined then they are not checked.
+ * compile FLAG `SCM_RECKLESS' is #defined then they are not checked.
  * Otherwise, number of argument checks for closures are made only when
  * the function position (whose value is the closure) of a combination is
  * not an ILOC or GLOC.  When the function position of a combination is a
  * symbol it will be checked only the first time it is evaluated because
  * it will then be replaced with an ILOC or GLOC.
  */
-#undef RECKLESS
-#define CAUTIOUS
+#undef SCM_RECKLESS
+#define SCM_CAUTIOUS
 
 /* After looking up a local for the first time, rewrite the
  * code graph, caching its position.
@@ -252,23 +252,13 @@ typedef long SCM_STACKITEM;
 #endif
 
 extern unsigned int scm_async_clock;
-#if 0
+
 #define SCM_ASYNC_TICK \
 { \
   if (0 == --scm_async_clock) \
     scm_async_click (); \
 } \
 
-#else
-#define SCM_ASYNC_TICK \
-{ \
-  if (0 == --scm_async_clock) \
-    scm_async_click (); \
-  SCM_THREAD_SWITCHING_CODE; \
-} \
-
-#endif
-
 #ifdef SCM_CAREFUL_INTS
 #define SCM_CHECK_NOT_DISABLED \
   if (scm_ints_disabled) \
@@ -326,6 +316,8 @@ extern unsigned int scm_async_clock;
 { \
   SCM_FENCE; \
   SCM_CHECK_NOT_ENABLED; \
+  SCM_THREAD_SWITCHING_CODE; \
+  SCM_FENCE; \
   scm_ints_disabled = 0; \
   SCM_FENCE; \
   SCM_THREAD_ALLOW; \
@@ -336,23 +328,72 @@ extern unsigned int scm_async_clock;
 
 #define SCM_REDEFER_INTS  \
 { \
+  SCM_FENCE; \
   SCM_THREAD_REDEFER; \
   ++scm_ints_disabled; \
+  SCM_FENCE; \
 } \
 
 
 #define SCM_REALLOW_INTS \
 { \
+  SCM_FENCE; \
   SCM_THREAD_REALLOW_1; \
+  SCM_THREAD_SWITCHING_CODE; \
+  SCM_FENCE; \
   --scm_ints_disabled; \
   if (!scm_ints_disabled) \
     { \
       SCM_THREAD_REALLOW_2; \
       SCM_ASYNC_TICK; \
     } \
+  SCM_FENCE; \
 } \
 
 
+#define SCM_TICK \
+{ \
+  SCM_DEFER_INTS; \
+  SCM_ALLOW_INTS; \
+} \
+
+
+
+/* Classification of critical sections
+ *
+ * When Guile moves to POSIX threads, it won't be possible to prevent
+ * context switching.  In fact, the whole idea of context switching is
+ * bogus if threads are run by different processors.  Therefore, we
+ * must ultimately eliminate all critical sections or enforce them by
+ * use of mutecis.
+ *
+ * All instances of SCM_DEFER_INTS and SCM_ALLOW_INTS should therefore
+ * be classified and replaced by one of the delimiters below.  If you
+ * understand what this is all about, I'd like to encourage you to
+ * help with this task.  The set of classes below must of course be
+ * incrementally augmented.
+ *
+ * MDJ 980419 <djurfeldt@nada.kth.se>
+ */
+
+/* A sections
+ *
+ * Allocation of a cell with type tag in the CAR.
+ *
+ * With POSIX threads, each thread will have a private pool of free
+ * cells.  Therefore, this type of section can be removed.  But!  It
+ * is important that the CDR is initialized first (with the CAR still
+ * indicating a free cell) so that we can guarantee a consistent heap
+ * at all times.
+ */
+
+#ifdef SCM_POSIX_THREADS
+#define SCM_ENTER_A_SECTION
+#define SCM_EXIT_A_SECTION
+#else
+#define SCM_ENTER_A_SECTION SCM_DEFER_INTS
+#define SCM_EXIT_A_SECTION SCM_ALLOW_INTS
+#endif
 
 
 
