@@ -1,4 +1,4 @@
-/*	Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+/*	Copyright (C) 1995, 1996, 1997, 1998 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -40,13 +40,19 @@
  * If you do not wish that, delete this exception notice.  */
 
 
-/* $Id: coop.c,v 1.2 1997/05/26 22:31:48 jimb Exp $ */
+/* $Id: coop.c,v 1.14 1998/10/13 23:17:09 jimb Exp $ */
 
 /* Cooperative thread library, based on QuickThreads */
 
-#include <qt.h>
+#ifdef HAVE_UNISTD_H 
+#include <unistd.h>
+#endif
 
-#define COOP_STKSIZE (0x10000)
+#include <qt.h>
+#include "eval.h"
+
+/* #define COOP_STKSIZE (0x10000) */
+#define COOP_STKSIZE (scm_eval_stack)
 
 /* `alignment' must be a power of 2. */
 #define COOP_STKALIGN(sp, alignment) \
@@ -69,14 +75,21 @@ coop_qinit (q)
 
   q->t.all_prev = NULL;
   q->t.all_next = NULL;
+#ifdef GUILE_ISELECT
+  q->t.nfds = 0;
+  q->t.readfds = NULL;
+  q->t.writefds = NULL;
+  q->t.exceptfds = NULL;
+  q->t.timeoutp = 0;
+#endif
 }
 
 
 #ifdef __STDC__
-static coop_t *
+coop_t *
 coop_qget (coop_q_t *q)
 #else
-static coop_t *
+coop_t *
 coop_qget (q)
      coop_q_t *q;
 #endif
@@ -96,10 +109,10 @@ coop_qget (q)
 
 
 #ifdef __STDC__
-static void
+void
 coop_qput (coop_q_t *q, coop_t *t)
 #else
-static void
+void
 coop_qput (q, t)
      coop_q_t *q;
      coop_t *t;
@@ -148,12 +161,12 @@ coop_all_qremove (q, t)
 
 /* Thread routines. */
 
-coop_q_t coop_global_runq;		 /* A queue of runable threads. */
-coop_q_t coop_global_sleepq;	 /* A queue of sleeping threads. */
-static coop_q_t tmp_queue;	         /* A temp working queue */
-coop_q_t coop_global_allq;  /* A queue of all threads. */
-static coop_t coop_global_main;    /* Thread for the process. */
-coop_t *coop_global_curr;	 /* Currently-executing thread. */
+coop_q_t coop_global_runq;	/* A queue of runable threads. */
+coop_q_t coop_global_sleepq;	/* A queue of sleeping threads. */
+coop_q_t coop_tmp_queue;        /* A temp working queue */
+coop_q_t coop_global_allq;      /* A queue of all threads. */
+static coop_t coop_global_main; /* Thread for the process. */
+coop_t *coop_global_curr;	/* Currently-executing thread. */
 
 static void *coop_starthelp (qt_t *old, void *ignore0, void *ignore1);
 static void coop_only (void *pu, void *pt, qt_userf_t *f);
@@ -171,7 +184,7 @@ coop_init()
 {
   coop_qinit (&coop_global_runq);
   coop_qinit (&coop_global_sleepq);
-  coop_qinit (&tmp_queue);
+  coop_qinit (&coop_tmp_queue);
   coop_qinit (&coop_global_allq);
   coop_global_curr = &coop_global_main;
 }
@@ -181,6 +194,7 @@ coop_init()
    and there are sleeping threads - wait until one wakes up. Otherwise,
    return NULL. */
 
+#ifndef GUILE_ISELECT
 #ifdef __STDC__
 coop_t *
 coop_next_runnable_thread()
@@ -204,9 +218,9 @@ coop_next_runnable_thread()
 	if (t->wakeup_time <= now)
 	  coop_qput(&coop_global_runq, t);
 	else
-	  coop_qput(&tmp_queue, t);
+	  coop_qput(&coop_tmp_queue, t);
       }
-    while ((t = coop_qget(&tmp_queue)) != NULL)
+    while ((t = coop_qget(&coop_tmp_queue)) != NULL)
       coop_qput(&coop_global_sleepq, t);
     
     t = coop_qget (&coop_global_runq);
@@ -215,7 +229,7 @@ coop_next_runnable_thread()
 
   return t;
 }
-
+#endif
 
 #ifdef __STDC__
 void
@@ -252,23 +266,24 @@ coop_starthelp (old, ignore0, ignore1)
 }
 
 #ifdef __STDC__
-void 
+int
 coop_mutex_init (coop_m *m)
 #else
-void 
+int
 coop_mutex_init (m)
      coop_m *m;
 #endif
 {
   m->owner = NULL;
   coop_qinit(&(m->waiting));
+  return 0;
 }
 
 #ifdef __STDC__
-void 
+int
 coop_mutex_lock (coop_m *m)
 #else
-void 
+int 
 coop_mutex_lock ()
      coop_m *m;
 #endif
@@ -284,19 +299,26 @@ coop_mutex_lock ()
       /* Record the current top-of-stack before going to sleep */
       coop_global_curr->top = &old;
 
+#ifdef GUILE_ISELECT
+      newthread = coop_wait_for_runnable_thread();
+      if (newthread == coop_global_curr)
+	coop_abort ();
+#else
       newthread = coop_next_runnable_thread();
+#endif
       old = coop_global_curr;
       coop_global_curr = newthread;
       QT_BLOCK (coop_yieldhelp, old, &(m->waiting), newthread->sp);
     }
+  return 0;
 }
 
 
 #ifdef __STDC__
-void 
+int 
 coop_mutex_unlock (coop_m *m)
 #else
-void 
+int 
 coop_mutex_unlock (m)
      coop_m *m;
 #endif
@@ -318,43 +340,83 @@ coop_mutex_unlock (m)
     {
       m->owner = NULL;
     }
+  return 0;
 }
 
 
 #ifdef __STDC__
-void 
+int 
+coop_mutex_destroy (coop_m *m)
+#else
+int 
+coop_mutex_destroy (m)
+     coop_m *m;
+#endif
+{
+  return 0;
+}
+
+
+#ifdef __STDC__
+int 
 coop_condition_variable_init (coop_c *c)
 #else
-void 
+int 
 coop_condition_variable_init (c)
      coop_c *c;
 #endif
 {
   coop_qinit(&(c->waiting));
+  return 0;
 }
 
 #ifdef __STDC__
-void 
+static int 
 coop_condition_variable_wait (coop_c *c)
 #else
-void 
+static int 
 coop_condition_variable_wait (c)
      coop_c *c;
 #endif
 {
   coop_t *old, *newthread;
 
+#ifdef GUILE_ISELECT
+  newthread = coop_wait_for_runnable_thread();
+  if (newthread == coop_global_curr)
+    coop_abort ();
+#else
   newthread = coop_next_runnable_thread();
+#endif
   old = coop_global_curr;
   coop_global_curr = newthread;
   QT_BLOCK (coop_yieldhelp, old, &(c->waiting), newthread->sp);
+  return 0;
 }
 
+
 #ifdef __STDC__
-void 
+int 
+coop_condition_variable_wait_mutex (coop_c *c, coop_m *m)
+#else
+int 
+coop_condition_variable_wait_mutex (c, m)
+     coop_c *c;
+     coop_m *m;
+#endif
+{
+  coop_mutex_unlock (m);
+  coop_condition_variable_wait (c);
+  coop_mutex_lock (m);
+  return 0;
+}
+
+
+#ifdef __STDC__
+int 
 coop_condition_variable_signal (coop_c *c)
 #else
-void 
+int 
 coop_condition_variable_signal (c)
      coop_c *c;
 #endif
@@ -365,6 +427,20 @@ coop_condition_variable_signal (c)
     {
       coop_qput (&coop_global_runq, newthread);
     }
+  return 0;
+}
+
+
+#ifdef __STDC__
+int 
+coop_condition_variable_destroy (coop_c *c)
+#else
+int 
+coop_condition_variable_destroy (c)
+     coop_c *c;
+#endif
+{
+  return 0;
 }
 
 
@@ -436,7 +512,15 @@ coop_abort ()
       free(coop_global_curr->joining);
     }
 
+#ifdef GUILE_ISELECT
+  scm_I_am_dead = 1;
+  do {
+    newthread = coop_wait_for_runnable_thread();
+  } while (newthread == coop_global_curr);
+  scm_I_am_dead = 0;
+#else
   newthread = coop_next_runnable_thread();
+#endif
   coop_all_qremove(&coop_global_allq, coop_global_curr);
   old = coop_global_curr;
   coop_global_curr = newthread;
@@ -490,7 +574,13 @@ coop_join()
       coop_qinit((coop_q_t *) t->joining);
     }
 
+#ifdef GUILE_ISELECT
+  newthread = coop_wait_for_runnable_thread();
+  if (newthread == coop_global_curr)
+    return;
+#else
   newthread = coop_next_runnable_thread();
+#endif
   old = coop_global_curr;
   coop_global_curr = newthread;
   QT_BLOCK (coop_yieldhelp, old, (coop_q_t *) t->joining, newthread->sp);
@@ -511,8 +601,13 @@ coop_yield()
 
   /* There may be no other runnable threads. Return if this is the 
      case. */
+#if GUILE_ISELECT
+  if (newthread == coop_global_curr)
+      return;
+#else
   if (newthread == NULL)
       return;
+#endif
 
   old = coop_global_curr;
 
@@ -540,16 +635,8 @@ coop_yieldhelp (sp, old, blockq)
 /* Replacement for the system's sleep() function. Does the right thing
    for the process - but not for the system (it busy-waits) */
 
-#ifdef __STDC__
-static void *
+void *
 coop_sleephelp (qt_t *sp, void *old, void *blockq)
-#else
-static void *
-coop_sleephelp (sp, old, bolckq)
-     qt_t *sp;
-     void *old;
-     void *blockq;
-#endif
 {
   ((coop_t *)old)->sp = sp;
   /* old is already on the sleep queue - so there's no need to
@@ -557,17 +644,39 @@ coop_sleephelp (sp, old, bolckq)
   return NULL;
 }
 
-#ifdef __STDC__
-unsigned
-sleep (unsigned s)
-#else
-unsigned
-sleep (s)
-     unsigned s;
-#endif
+#ifdef GUILE_ISELECT
+
+unsigned long 
+scm_thread_usleep (unsigned long usec)
+{
+  struct timeval timeout;
+  timeout.tv_sec = 0;
+  timeout.tv_usec = usec;
+  scm_internal_select (0, NULL, NULL, NULL, &timeout);
+  return 0;  /* Maybe we should calculate actual time slept,
+		but this is faster... :) */
+}
+
+unsigned long
+scm_thread_sleep (unsigned long sec)
+{
+  time_t now = time (NULL);
+  struct timeval timeout;
+  unsigned long slept;
+  timeout.tv_sec = sec;
+  timeout.tv_usec = 0;
+  scm_internal_select (0, NULL, NULL, NULL, &timeout);
+  slept = time (NULL) - now;
+  return slept > sec ? 0 : sec - slept;
+}
+
+#else /* GUILE_ISELECT */
+
+unsigned long
+scm_thread_sleep (unsigned long s)
 {
   coop_t *newthread, *old;
-  time_t now = time(NULL);
+  time_t now = time (NULL);
   coop_global_curr->wakeup_time = now + s;
 
   /* Put the current thread on the sleep queue */
@@ -586,3 +695,15 @@ sleep (s)
 
   return s;
 }
+
+unsigned long 
+scm_thread_usleep (unsigned long usec)
+{
+  /* We're so cheap.  */
+  scm_thread_sleep (usec / 1000000);
+  struct timeval timeout;
+  return 0;  /* Maybe we should calculate actual time slept,
+		but this is faster... :) */
+}
+
+#endif /* GUILE_ISELECT */
