@@ -1,4 +1,4 @@
-/*	Copyright (C) 1995,1996,1997 Free Software Foundation, Inc.
+/*	Copyright (C) 1995,1996,1997,1998 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -77,9 +77,7 @@ char *alloca ();
 #include <stdio.h>
 #include "_scm.h"
 #include "debug.h"
-#include "append.h"
 #include "alist.h"
-#include "sequences.h"
 #include "eq.h"
 #include "continuations.h"
 #include "throw.h"
@@ -88,13 +86,10 @@ char *alloca ();
 #include "procprop.h"
 #include "hashtab.h"
 #include "hash.h"
-
-#ifdef DEBUG_EXTENSIONS
-#include "debug.h"
-#endif /* DEBUG_EXTENSIONS */
-
 #include "srcprop.h"
 #include "stackchk.h"
+#include "objects.h"
+#include "feature.h"
 
 #include "eval.h"
 
@@ -259,10 +254,8 @@ scm_ilookup (iloc, env)
 
 #endif /* USE_THREADS */
 
-SCM *
-scm_lookupcar1 (vloc, genv)
-     SCM vloc;
-     SCM genv;
+static SCM *
+scm_lookupcar1 (SCM vloc, SCM genv)
 {
   SCM env = genv;
   register SCM *al, fl, var = SCM_CAR (vloc);
@@ -280,6 +273,7 @@ scm_lookupcar1 (vloc, genv)
       for (fl = SCM_CAR (*al); SCM_NIMP (fl); fl = SCM_CDR (fl))
 	{
 	  if (SCM_NCONSP (fl))
+	    {
 	      if (fl == var)
 	      {
 #ifdef MEMOIZE_LOCALS
@@ -291,13 +285,14 @@ scm_lookupcar1 (vloc, genv)
 #endif
 		return SCM_CDRLOC (*al);
 	      }
-	    else
-	      break;
+	      else
+		break;
+	    }
 	  al = SCM_CDRLOC (*al);
 	  if (SCM_CAR (fl) == var)
 	    {
 #ifdef MEMOIZE_LOCALS
-#ifndef RECKLESS		/* letrec inits to SCM_UNDEFINED */
+#ifndef SCM_RECKLESS		/* letrec inits to SCM_UNDEFINED */
 	      if (SCM_UNBNDP (SCM_CAR (*al)))
 		{
 		  env = SCM_EOL;
@@ -335,7 +330,7 @@ scm_lookupcar1 (vloc, genv)
     else
       var = vcell;
   }
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
   if (SCM_NNULLP (env) || SCM_UNBNDP (SCM_CDR (var)))
     {
       var = SCM_CAR (var);
@@ -652,23 +647,23 @@ scm_m_lambda (xorig, env)
   if (scm_ilength (x) < 2)
     goto badforms;
   proc = SCM_CAR (x);
-  if SCM_NULLP
-    (proc) goto memlambda;
-  if SCM_IMP
-    (proc) goto badforms;
-  if SCM_SYMBOLP
-    (proc) goto memlambda;
-  if SCM_NCONSP
-    (proc) goto badforms;
-  while SCM_NIMP
-    (proc)
+  if (SCM_NULLP (proc))
+    goto memlambda;
+  if (SCM_IMP (proc))
+    goto badforms;
+  if (SCM_SYMBOLP (proc))
+    goto memlambda;
+  if (SCM_NCONSP (proc))
+    goto badforms;
+  while (SCM_NIMP (proc))
     {
-      if SCM_NCONSP
-	(proc)
+      if (SCM_NCONSP (proc))
+	{
 	  if (!SCM_SYMBOLP (proc))
-	  goto badforms;
-	else
-	  goto memlambda;
+	    goto badforms;
+	  else
+	    goto memlambda;
+	}
       if (!(SCM_NIMP (SCM_CAR (proc)) && SCM_SYMBOLP (SCM_CAR (proc))))
 	goto badforms;
       proc = SCM_CDR (proc);
@@ -879,12 +874,25 @@ scm_m_define (x, env)
     {
       x = evalcar (x, env);
 #ifdef DEBUG_EXTENSIONS
-      if (SCM_REC_PROCNAMES_P && SCM_NIMP (x) && SCM_CLOSUREP (x))
-	scm_set_procedure_property_x (x, scm_i_name, proc);
+      if (SCM_REC_PROCNAMES_P && SCM_NIMP (x))
+	{
+	  arg1 = x;
+	proc:
+	  if (SCM_CLOSUREP (arg1)
+	      /* Only the first definition determines the name. */
+	      && scm_procedure_property (arg1, scm_i_name) == SCM_BOOL_F)
+	    scm_set_procedure_property_x (arg1, scm_i_name, proc);
+	  else if (SCM_TYP16 (arg1) == scm_tc16_macro
+		   && SCM_CDR (arg1) != arg1)
+	    {
+	      arg1 = SCM_CDR (arg1);
+	      goto proc;
+	    }
+	}
 #endif
       arg1 = scm_sym2vcell (proc, env_top_level (env), SCM_BOOL_T);
 #if 0
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
       if (SCM_NIMP (SCM_CDR (arg1)) && ((SCM) SCM_SNAME (SCM_CDR (arg1)) == proc)
 	  && (SCM_CDR (arg1) != x))
 	scm_warn ("redefining built-in ", SCM_CHARS (proc));
@@ -918,7 +926,7 @@ scm_m_undefine (x, env)
   ASSYNT (SCM_NFALSEP (arg1) && !SCM_UNBNDP (SCM_CDR (arg1)),
 	  x, "variable already unbound ", s_undefine);
 #if 0
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
   if (SCM_NIMP (SCM_CDR (arg1)) && ((SCM) SCM_SNAME (SCM_CDR (arg1)) == x))
     scm_warn ("undefining built-in ", SCM_CHARS (x));
   else
@@ -1241,7 +1249,7 @@ scm_unmemocopy (x, env)
     return unmemocopy (x, env);
 }
 
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
 
 int 
 scm_badargsp (formals, args)
@@ -1268,50 +1276,46 @@ long scm_tc16_macro;
 
 
 SCM 
-scm_eval_args (l, env)
+scm_eval_args (l, env, proc)
      SCM l;
      SCM env;
+     SCM proc;
 {
-  SCM res = SCM_EOL, *lloc = &res;
+  SCM results = SCM_EOL, *lloc = &results, res;
   while (SCM_NIMP (l))
     {
-      *lloc = scm_cons (EVALCAR (l, env), SCM_EOL);
+#ifdef SCM_CAUTIOUS
+      if (SCM_IMP (l))
+	goto wrongnumargs;
+      else if (SCM_CONSP (l))
+	{
+	  if (SCM_IMP (SCM_CAR (l)))
+	    res = EVALIM (SCM_CAR (l), env);
+	  else
+	    res = EVALCELLCAR (l, env);
+	}
+      else if (SCM_TYP3 (l) == 1)
+	{
+	  if ((res = SCM_GLOC_VAL (SCM_CAR (l))) == 0)
+	    res = SCM_CAR (l); /* struct planted in code */
+	}
+      else
+	goto wrongnumargs;
+#else
+      res = EVALCAR (l, env);
+#endif
+      *lloc = scm_cons (res, SCM_EOL);
       lloc = SCM_CDRLOC (*lloc);
       l = SCM_CDR (l);
     }
-  return res;
-}
-
-
-/* The SCM_CEVAL and SCM_APPLY functions use this function instead of
-   calling setjmp directly, to make sure that local variables don't
-   have their values clobbered by a longjmp.
-
-   According to Harbison & Steele, "Automatic variables local to the
-   function containing setjmp are guaranteed to have their correct
-   value in ANSI C only if they have a volatile-qualified type or if
-   their values were not changed between the original call to setjmp
-   and the corresponding longjmp call."
-
-   SCM_CEVAL and SCM_APPLY are too complex for me to see how to meet
-   the second condition, and making x and env volatile would be a
-   speed problem, so we'll just trivially meet the first, by having no
-   "automatic variables local to the function containing setjmp."  */
-
-/* Actually, this entire approach is bogus, because setjmp ends up
-   capturing the stack frame of the wrapper function, which then
-   returns, rendering the jump buffer invalid.  Duh.  Gotta find a
-   better way...  -JimB  */
-#define safe_setjmp(x) setjmp (x)
-static int
-unsafe_setjmp (jmp_buf env)
-{
-  /* I think ANSI requires us to write the function this way, instead
-     of just saying "return setjmp (env)".  Maybe I'm being silly.
-     See Harbison & Steele, third edition, p. 353.  */
-  int val;
-  val = setjmp (env);
-  return val;
+#ifdef SCM_CAUTIOUS
+  if (SCM_NNULLP (l))
+    {
+    wrongnumargs:
+      scm_wrong_num_args (proc);
+    }
+#endif
+  return results;
 }
 
 
@@ -1348,23 +1352,20 @@ unsafe_setjmp (jmp_buf env)
 #define ENTER_APPLY \
 {\
   SCM_SET_ARGSREADY (debug);\
-  if (CHECK_APPLY)\
+  if (CHECK_APPLY && SCM_TRAPS_P)\
     if (SCM_APPLY_FRAME_P || (SCM_TRACE_P && PROCTRACEP (proc)))\
       {\
 	SCM tmp, tail = SCM_TRACED_FRAME_P (debug) ? SCM_BOOL_T : SCM_BOOL_F;\
 	SCM_SET_TRACED_FRAME (debug); \
-	SCM_APPLY_FRAME_P = 0; \
-	SCM_TRACE_P = 0; \
-	SCM_RESET_DEBUG_MODE; \
 	if (SCM_CHEAPTRAPS_P)\
 	  {\
 	    tmp = scm_make_debugobj (&debug);\
 	    scm_ithrow (scm_i_apply_frame, scm_cons2 (tmp, tail, SCM_EOL), 0);\
-	  }\
+ 	  }\
 	else\
 	  {\
 	    scm_make_cont (&tmp);\
-	    if (!safe_setjmp (SCM_JMPBUF (tmp)))\
+	    if (!setjmp (SCM_JMPBUF (tmp)))\
 	      scm_ithrow (scm_i_apply_frame, scm_cons2 (tmp, tail, SCM_EOL), 0);\
 	  }\
       }\
@@ -1405,6 +1406,12 @@ int scm_debug_eframe_size;
 
 int scm_debug_mode, scm_check_entry_p, scm_check_apply_p, scm_check_exit_p;
 
+int scm_eval_stack;
+
+scm_option scm_eval_opts[] = {
+  { SCM_OPTION_INTEGER, "stack", 22000, "Size of thread stacks (in machine words)." }
+};
+
 scm_option scm_debug_opts[] = {
   { SCM_OPTION_BOOLEAN, "cheap", 1,
     "*Flyweight representation of the stack at traps." },
@@ -1422,27 +1429,88 @@ scm_option scm_debug_opts[] = {
   { SCM_OPTION_INTEGER, "depth", 20, "Maximal length of printed backtrace." },
   { SCM_OPTION_BOOLEAN, "backtrace", 0, "Show backtrace on error." },
   { SCM_OPTION_BOOLEAN, "debug", 0, "Use the debugging evaluator." },
-  { SCM_OPTION_INTEGER, "stack", 20000, "Stack size limit (0 = no check)." }
+  { SCM_OPTION_INTEGER, "stack", 20000, "Stack size limit (measured in words; 0 = no check)." }
 };
 
 scm_option scm_evaluator_trap_table[] = {
+  { SCM_OPTION_BOOLEAN, "traps", 0, "Enable evaluator traps." },
   { SCM_OPTION_BOOLEAN, "enter-frame", 0, "Trap when eval enters new frame." },
   { SCM_OPTION_BOOLEAN, "apply-frame", 0, "Trap when entering apply." },
   { SCM_OPTION_BOOLEAN, "exit-frame", 0, "Trap when exiting eval or apply." }
 };
 
+SCM_PROC (s_eval_options_interface, "eval-options-interface", 0, 1, 0, scm_eval_options_interface);
+
 SCM
-scm_deval_args (l, env, lloc)
-     SCM l, env, *lloc;
+scm_eval_options_interface (SCM setting)
 {
-  SCM *res = lloc;
+  SCM ans;
+  SCM_DEFER_INTS;
+  ans = scm_options (setting,
+		     scm_eval_opts,
+		     SCM_N_EVAL_OPTIONS,
+		     s_eval_options_interface);
+  scm_eval_stack = SCM_EVAL_STACK * sizeof (void *);
+  SCM_ALLOW_INTS;
+  return ans;
+}
+
+SCM_PROC (s_evaluator_traps, "evaluator-traps-interface", 0, 1, 0, scm_evaluator_traps);
+
+SCM
+scm_evaluator_traps (setting)
+     SCM setting;
+{
+  SCM ans;
+  SCM_DEFER_INTS;
+  ans = scm_options (setting,
+		     scm_evaluator_trap_table,
+		     SCM_N_EVALUATOR_TRAPS,
+		     s_evaluator_traps);
+  SCM_RESET_DEBUG_MODE;
+  SCM_ALLOW_INTS
+  return ans;
+}
+
+SCM
+scm_deval_args (l, env, proc, lloc)
+     SCM l, env, proc, *lloc;
+{
+  SCM *results = lloc, res;
   while (SCM_NIMP (l))
     {
-      *lloc = scm_cons (EVALCAR (l, env), SCM_EOL);
+#ifdef SCM_CAUTIOUS
+      if (SCM_IMP (l))
+	goto wrongnumargs;
+      else if (SCM_CONSP (l))
+	{
+	  if (SCM_IMP (SCM_CAR (l)))
+	    res = EVALIM (SCM_CAR (l), env);
+	  else
+	    res = EVALCELLCAR (l, env);
+	}
+      else if (SCM_TYP3 (l) == 1)
+	{
+	  if ((res = SCM_GLOC_VAL (SCM_CAR (l))) == 0)
+	    res = SCM_CAR (l); /* struct planted in code */
+	}
+      else
+	goto wrongnumargs;
+#else
+      res = EVALCAR (l, env);
+#endif
+      *lloc = scm_cons (res, SCM_EOL);
       lloc = SCM_CDRLOC (*lloc);
       l = SCM_CDR (l);
     }
-  return *res;
+#ifdef SCM_CAUTIOUS
+  if (SCM_NNULLP (l))
+    {
+    wrongnumargs:
+      scm_wrong_num_args (proc);
+    }
+#endif
+  return *results;
 }
 
 #endif /* !DEVAL */
@@ -1533,19 +1601,17 @@ loop:
 start:
   debug.info->e.exp = x;
   debug.info->e.env = env;
-  if (CHECK_ENTRY)
+  if (CHECK_ENTRY && SCM_TRAPS_P)
     if (SCM_ENTER_FRAME_P || (SCM_BREAKPOINTS_P && SRCBRKP (x)))
       {
 	SCM tail = SCM_TAILRECP (debug) ? SCM_BOOL_T : SCM_BOOL_F;
 	SCM_SET_TAILREC (debug);
-	SCM_ENTER_FRAME_P = 0;
-	SCM_RESET_DEBUG_MODE;
 	if (SCM_CHEAPTRAPS_P)
 	  t.arg1 = scm_make_debugobj (&debug);
 	else
 	  {
 	    scm_make_cont (&t.arg1);
-	    if (safe_setjmp (SCM_JMPBUF (t.arg1)))
+	    if (setjmp (SCM_JMPBUF (t.arg1)))
 	      {
 		x = SCM_THROW_VALUE (t.arg1);
 		if (SCM_IMP (x))
@@ -1567,7 +1633,7 @@ start:
 #if defined (USE_THREADS) || defined (DEVAL)
 dispatch:
 #endif
-  SCM_ASYNC_TICK;
+  SCM_TICK;
   switch (SCM_TYP7 (x))
     {
     case scm_tcs_symbols:
@@ -1818,8 +1884,22 @@ dispatch:
       x = SCM_CDR (x);
       x = evalcar (x, env);
 #ifdef DEBUG_EXTENSIONS
-      if (SCM_REC_PROCNAMES_P && SCM_NIMP (x) && SCM_CLOSUREP (x))
-	scm_set_procedure_property_x (x, scm_i_name, proc);
+      if (SCM_REC_PROCNAMES_P && SCM_NIMP (x))
+	{
+	  t.arg1 = x;
+	proc:
+	  if (SCM_CLOSUREP (t.arg1)
+	      /* Only the first definition determines the name. */
+	      && (scm_procedure_property (t.arg1, scm_i_inner_name)
+		  == SCM_BOOL_F))
+	    scm_set_procedure_property_x (t.arg1, scm_i_inner_name, proc);
+	  else if (SCM_TYP16 (t.arg1) == scm_tc16_macro
+		   && SCM_CDR (t.arg1) != t.arg1)
+	    {
+	      t.arg1 = SCM_CDR (t.arg1);
+	      goto proc;
+	    }
+	}
 #endif
       env = SCM_CAR (env);
       SCM_DEFER_INTS;
@@ -1854,17 +1934,34 @@ dispatch:
 	  SCM_ASRTGO (SCM_NIMP (proc), badfun);
 	  if (SCM_CLOSUREP (proc))
 	    {
+	      SCM argl, tl;
 	      PREP_APPLY (proc, SCM_EOL);
 	      t.arg1 = SCM_CDR (SCM_CDR (x));
 	      t.arg1 = EVALCAR (t.arg1, env);
 #ifdef DEVAL
 	      debug.info->a.args = t.arg1;
 #endif
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
 	      if (scm_badargsp (SCM_CAR (SCM_CODE (proc)), t.arg1))
 		goto wrongnumargs;
 #endif
-	      env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), t.arg1, SCM_ENV (proc));
+	      /* Copy argument list */
+	      if (SCM_IMP (t.arg1))
+		argl = t.arg1;
+	      else
+		{
+		  argl = tl = scm_cons (SCM_CAR (t.arg1), SCM_UNSPECIFIED);
+		  while (SCM_NIMP (t.arg1 = SCM_CDR (t.arg1))
+			 && SCM_CONSP (t.arg1))
+		    {
+		      SCM_SETCDR (tl, scm_cons (SCM_CAR (t.arg1),
+						SCM_UNSPECIFIED));
+		      tl = SCM_CDR (tl);
+		    }
+		  SCM_SETCDR (tl, t.arg1);
+		}
+	      
+	      env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), argl, SCM_ENV (proc));
 	      x = SCM_CODE (proc);
 	      goto cdrxbegin;
 	    }
@@ -1873,7 +1970,7 @@ dispatch:
 
 	case (SCM_ISYMNUM (SCM_IM_CONT)):
 	  scm_make_cont (&t.arg1);
-	  if (safe_setjmp (SCM_JMPBUF (t.arg1)))
+	  if (setjmp (SCM_JMPBUF (t.arg1)))
 	    {
 	      SCM val;
 	      val = SCM_THROW_VALUE (t.arg1);
@@ -1911,9 +2008,7 @@ dispatch:
     case scm_tc7_llvect:
 #endif
     case scm_tc7_string:
-    case scm_tc7_mb_string:
     case scm_tc7_substring:
-    case scm_tc7_mb_substring:
     case scm_tc7_smob:
     case scm_tcs_closures:
     case scm_tcs_subrs:
@@ -1923,8 +2018,8 @@ dispatch:
     case (127 & SCM_ILOC00):
       proc = *scm_ilookup (SCM_CAR (x), env);
       SCM_ASRTGO (SCM_NIMP (proc), badfun);
-#ifndef RECKLESS
-#ifdef CAUTIOUS
+#ifndef SCM_RECKLESS
+#ifdef SCM_CAUTIOUS
       goto checkargs;
 #endif
 #endif
@@ -1935,8 +2030,8 @@ dispatch:
     case scm_tcs_cons_gloc:
       proc = SCM_GLOC_VAL (SCM_CAR (x));
       SCM_ASRTGO (SCM_NIMP (proc), badfun);
-#ifndef RECKLESS
-#ifdef CAUTIOUS
+#ifndef SCM_RECKLESS
+#ifdef SCM_CAUTIOUS
       goto checkargs;
 #endif
 #endif
@@ -2016,8 +2111,8 @@ dispatch:
       else
 	proc = SCM_CEVAL (SCM_CAR (x), env);
       SCM_ASRTGO (SCM_NIMP (proc), badfun);
-#ifndef RECKLESS
-#ifdef CAUTIOUS
+#ifndef SCM_RECKLESS
+#ifdef SCM_CAUTIOUS
     checkargs:
 #endif
       if (SCM_CLOSUREP (proc))
@@ -2072,6 +2167,29 @@ evapply:
 	x = SCM_CODE (proc);
 	env = EXTEND_ENV (SCM_CAR (x), SCM_EOL, SCM_ENV (proc));
 	goto cdrxbegin;
+      case scm_tcs_cons_gloc:
+	if (SCM_I_OPERATORP (proc))
+	  {
+	    x = (SCM_I_ENTITYP (proc)
+		 ? SCM_ENTITY_PROC_0 (proc)
+		 : SCM_OPERATOR_PROC_0 (proc));
+	    if (SCM_NIMP (x))
+	      {
+		if (SCM_TYP7 (x) == scm_tc7_subr_1)
+		  RETURN (SCM_SUBRF (x) (proc))
+		else if (SCM_CLOSUREP (x))
+		  {
+		    t.arg1 = proc;
+		    proc = x;
+#ifdef DEVAL
+		    debug.info->a.args = scm_cons (t.arg1, SCM_EOL);
+		    debug.info->a.proc = proc;
+#endif
+		    goto clos1;
+		  }
+	      }
+	    /* Fall through. */
+	  }
       case scm_tc7_contin:
       case scm_tc7_subr_1:
       case scm_tc7_subr_2:
@@ -2092,11 +2210,26 @@ evapply:
 
   /* must handle macros by here */
   x = SCM_CDR (x);
-#ifdef CAUTIOUS
+#ifdef SCM_CAUTIOUS
   if (SCM_IMP (x))
     goto wrongnumargs;
-#endif
+  else if (SCM_CONSP (x))
+    {
+      if (SCM_IMP (SCM_CAR (x)))
+	t.arg1 = EVALIM (SCM_CAR (x), env);
+      else
+	t.arg1 = EVALCELLCAR (x, env);
+    }
+  else if (SCM_TYP3 (x) == 1)
+    {
+      if ((t.arg1 = SCM_GLOC_VAL (SCM_CAR (x))) == 0)
+	t.arg1 = SCM_CAR (x); /* struct planted in code */
+    }
+  else
+    goto wrongnumargs;
+#else
   t.arg1 = EVALCAR (x, env);
+#endif
 #ifdef DEVAL
   debug.info->a.args = scm_cons (t.arg1, SCM_EOL);
 #endif
@@ -2169,6 +2302,7 @@ evapply:
 	  goto evap2;
 #endif
 	case scm_tcs_closures:
+	clos1:
 	  x = SCM_CODE (proc);
 #ifdef DEVAL
 	  env = EXTEND_ENV (SCM_CAR (x), debug.info->a.args, SCM_ENV (proc));
@@ -2178,6 +2312,31 @@ evapply:
 	  goto cdrxbegin;
 	case scm_tc7_contin:
 	  scm_call_continuation (proc, t.arg1);
+	case scm_tcs_cons_gloc:
+	  if (SCM_I_OPERATORP (proc))
+	    {
+	      x = (SCM_I_ENTITYP (proc)
+		   ? SCM_ENTITY_PROC_1 (proc)
+		   : SCM_OPERATOR_PROC_1 (proc));
+	      if (SCM_NIMP (x))
+		{
+		  if (SCM_TYP7 (x) == scm_tc7_subr_2)
+		    RETURN (SCM_SUBRF (x) (proc, t.arg1))
+		  else if (SCM_CLOSUREP (x))
+		    {
+		      arg2 = t.arg1;
+		      t.arg1 = proc;
+		      proc = x;
+#ifdef DEVAL
+		      debug.info->a.args = scm_cons (t.arg1,
+						     debug.info->a.args);
+		      debug.info->a.proc = proc;
+#endif
+		      goto clos2;
+		    }
+		}
+	      /* Fall through. */
+	    }
 	case scm_tc7_subr_2:
 	case scm_tc7_subr_0:
 	case scm_tc7_subr_3:
@@ -2187,12 +2346,27 @@ evapply:
 	  goto badfun;
 	}
     }
-#ifdef CAUTIOUS
+#ifdef SCM_CAUTIOUS
   if (SCM_IMP (x))
     goto wrongnumargs;
+  else if (SCM_CONSP (x))
+    {
+      if (SCM_IMP (SCM_CAR (x)))
+	arg2 = EVALIM (SCM_CAR (x), env);
+      else
+	arg2 = EVALCELLCAR (x, env);
+    }
+  else if (SCM_TYP3 (x) == 1)
+    {
+      if ((arg2 = SCM_GLOC_VAL (SCM_CAR (x))) == 0)
+	arg2 = SCM_CAR (x); /* struct planted in code */
+    }
+  else
+    goto wrongnumargs;
+#else
+  arg2 = EVALCAR (x, env);
 #endif
   {				/* have two or more arguments */
-    arg2 = EVALCAR (x, env);
 #ifdef DEVAL
     debug.info->a.args = scm_cons2 (t.arg1, arg2, SCM_EOL);
 #endif
@@ -2227,7 +2401,8 @@ evapply:
 #else
 	  RETURN (SCM_APPLY (SCM_CCLO_SUBR (proc), proc,
 			     scm_cons2 (t.arg1, arg2,
-					scm_cons (scm_eval_args (x, env), SCM_EOL))));
+					scm_cons (scm_eval_args (x, env, proc),
+						  SCM_EOL))));
 #endif
 	  /*    case scm_tc7_cclo:
 		x = scm_cons(arg2, scm_eval_args(x, env));
@@ -2236,6 +2411,34 @@ evapply:
 		proc = SCM_CCLO_SUBR(proc);
 		goto evap3; */
 #endif
+	case scm_tcs_cons_gloc:
+	  if (SCM_I_OPERATORP (proc))
+	    {
+	      x = (SCM_I_ENTITYP (proc)
+		   ? SCM_ENTITY_PROC_2 (proc)
+		   : SCM_OPERATOR_PROC_2 (proc));
+	      if (SCM_NIMP (x))
+		{
+		  if (SCM_TYP7 (x) == scm_tc7_subr_3)
+		    RETURN (SCM_SUBRF (x) (proc, t.arg1, arg2))
+		  else if (SCM_CLOSUREP (x))
+		    {
+#ifdef DEVAL
+		      SCM_SET_ARGSREADY (debug);
+		      debug.info->a.args = scm_cons (proc,
+						     debug.info->a.args);
+		      debug.info->a.proc = x;
+#endif
+		      env = EXTEND_ENV (SCM_CAR (SCM_CODE (x)),
+					scm_cons2 (proc, t.arg1,
+						   scm_cons (arg2, SCM_EOL)),
+					SCM_ENV (x));
+		      x = SCM_CODE (x);
+		      goto cdrxbegin;
+		    }
+		}
+	      /* Fall through. */
+	    }
 	case scm_tc7_subr_0:
 	case scm_tc7_cxr:
 	case scm_tc7_subr_1o:
@@ -2246,18 +2449,27 @@ evapply:
 	default:
 	  goto badfun;
 	case scm_tcs_closures:
+	clos2:
 #ifdef DEVAL
-	  env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), debug.info->a.args, SCM_ENV (proc));
+	  env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)),
+			    debug.info->a.args,
+			    SCM_ENV (proc));
 #else
-	  env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), scm_cons2 (t.arg1, arg2, SCM_EOL), SCM_ENV (proc));
+	  env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)),
+			    scm_cons2 (t.arg1, arg2, SCM_EOL), SCM_ENV (proc));
 #endif
 	  x = SCM_CODE (proc);
 	  goto cdrxbegin;
 	}
     }
+#ifdef SCM_CAUTIOUS
+    if (SCM_IMP (x) || SCM_NECONSP (x))
+      goto wrongnumargs;
+#endif
 #ifdef DEVAL
     debug.info->a.args = scm_cons2 (t.arg1, arg2,
-	    scm_deval_args (x, env, SCM_CDRLOC (SCM_CDR (debug.info->a.args))));
+      scm_deval_args (x, env, proc,
+		      SCM_CDRLOC (SCM_CDR (debug.info->a.args))));
 #endif
     ENTER_APPLY;
     switch (SCM_TYP7 (proc))
@@ -2265,15 +2477,18 @@ evapply:
 #ifdef DEVAL
       case scm_tc7_subr_3:
 	SCM_ASRTGO (SCM_NULLP (SCM_CDR (x)), wrongnumargs);
-	RETURN (SCM_SUBRF (proc) (t.arg1, arg2, SCM_CAR (SCM_CDR (SCM_CDR (debug.info->a.args)))));
+	RETURN (SCM_SUBRF (proc) (t.arg1, arg2,
+				  SCM_CADDR (debug.info->a.args)));
       case scm_tc7_asubr:
 #ifdef BUILTIN_RPASUBR
 	t.arg1 = SCM_SUBRF(proc)(t.arg1, arg2);
 	arg2 = SCM_CDR (SCM_CDR (debug.info->a.args));
-	do {
-	  t.arg1 = SCM_SUBRF(proc)(t.arg1, SCM_CAR (arg2));
-	  arg2 = SCM_CDR (arg2);
-	} while (SCM_NIMP (arg2));
+	do
+	  {
+	    t.arg1 = SCM_SUBRF(proc)(t.arg1, SCM_CAR (arg2));
+	    arg2 = SCM_CDR (arg2);
+	  }
+	while (SCM_NIMP (arg2));
 	RETURN (t.arg1)
 #endif /* BUILTIN_RPASUBR */
       case scm_tc7_rpsubr:
@@ -2281,18 +2496,24 @@ evapply:
 	if (SCM_FALSEP (SCM_SUBRF (proc) (t.arg1, arg2)))
 	  RETURN (SCM_BOOL_F)
 	t.arg1 = SCM_CDR (SCM_CDR (debug.info->a.args));
-	do {
-	  if (SCM_FALSEP (SCM_SUBRF (proc) (arg2, SCM_CAR (t.arg1))))
-	    RETURN (SCM_BOOL_F)
-	  arg2 = SCM_CAR (t.arg1);
-	  t.arg1 = SCM_CDR (t.arg1);
-	} while (SCM_NIMP (t.arg1));
+	do
+	  {
+	    if (SCM_FALSEP (SCM_SUBRF (proc) (arg2, SCM_CAR (t.arg1))))
+	      RETURN (SCM_BOOL_F)
+		arg2 = SCM_CAR (t.arg1);
+	    t.arg1 = SCM_CDR (t.arg1);
+	  }
+	while (SCM_NIMP (t.arg1));
 	RETURN (SCM_BOOL_T)
 #else /* BUILTIN_RPASUBR */
-	RETURN (SCM_APPLY (proc, t.arg1, scm_acons (arg2, SCM_CDR (SCM_CDR (debug.info->a.args)), SCM_EOL)))
+	RETURN (SCM_APPLY (proc, t.arg1,
+			   scm_acons (arg2,
+				      SCM_CDR (SCM_CDR (debug.info->a.args)),
+				      SCM_EOL)))
 #endif /* BUILTIN_RPASUBR */
       case scm_tc7_lsubr_2:
-	RETURN (SCM_SUBRF (proc) (t.arg1, arg2, SCM_CDR (SCM_CDR (debug.info->a.args))))
+	RETURN (SCM_SUBRF (proc) (t.arg1, arg2,
+				  SCM_CDR (SCM_CDR (debug.info->a.args))))
       case scm_tc7_lsubr:
 	RETURN (SCM_SUBRF (proc) (debug.info->a.args))
 #ifdef CCLO
@@ -2312,32 +2533,41 @@ evapply:
 	RETURN (SCM_SUBRF (proc) (t.arg1, arg2, EVALCAR (x, env)));
       case scm_tc7_asubr:
 #ifdef BUILTIN_RPASUBR
-	t.arg1 = SCM_SUBRF(proc)(t.arg1, arg2);
-	do {
-	  t.arg1 = SCM_SUBRF(proc)(t.arg1, EVALCAR(x, env));
-	  x = SCM_CDR(x);
-	} while (SCM_NIMP (x));
+	t.arg1 = SCM_SUBRF (proc) (t.arg1, arg2);
+	do
+	  {
+	    t.arg1 = SCM_SUBRF(proc)(t.arg1, EVALCAR(x, env));
+	    x = SCM_CDR(x);
+	  }
+	while (SCM_NIMP (x));
 	RETURN (t.arg1)
 #endif /* BUILTIN_RPASUBR */
       case scm_tc7_rpsubr:
 #ifdef BUILTIN_RPASUBR
 	if (SCM_FALSEP (SCM_SUBRF (proc) (t.arg1, arg2)))
 	  RETURN (SCM_BOOL_F)
-	do {
-	  t.arg1 = EVALCAR (x, env);
-	  if (SCM_FALSEP (SCM_SUBRF (proc) (arg2, t.arg1)))
-	    RETURN (SCM_BOOL_F)
-	  arg2 = t.arg1;
-	  x = SCM_CDR (x);
-	} while (SCM_NIMP (x));
+	do
+	  {
+	    t.arg1 = EVALCAR (x, env);
+	    if (SCM_FALSEP (SCM_SUBRF (proc) (arg2, t.arg1)))
+	      RETURN (SCM_BOOL_F)
+		arg2 = t.arg1;
+	    x = SCM_CDR (x);
+	  }
+	while (SCM_NIMP (x));
 	RETURN (SCM_BOOL_T)
 #else /* BUILTIN_RPASUBR */
-	RETURN (SCM_APPLY (proc, t.arg1, scm_acons (arg2, scm_eval_args (x, env), SCM_EOL)));
+	RETURN (SCM_APPLY (proc, t.arg1,
+			   scm_acons (arg2,
+				      scm_eval_args (x, env, proc),
+				      SCM_EOL)));
 #endif /* BUILTIN_RPASUBR */
       case scm_tc7_lsubr_2:
-	RETURN (SCM_SUBRF (proc) (t.arg1, arg2, scm_eval_args (x, env)));
+	RETURN (SCM_SUBRF (proc) (t.arg1, arg2, scm_eval_args (x, env, proc)));
       case scm_tc7_lsubr:
-	RETURN (SCM_SUBRF (proc) (scm_cons2 (t.arg1, arg2, scm_eval_args (x, env))));
+	RETURN (SCM_SUBRF (proc) (scm_cons2 (t.arg1,
+					     arg2,
+					     scm_eval_args (x, env, proc))));
 #ifdef CCLO
       case scm_tc7_cclo:
 	goto cclon;
@@ -2347,11 +2577,54 @@ evapply:
 	SCM_SET_ARGSREADY (debug);
 #endif
 	env = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)),
-			      scm_cons2 (t.arg1, arg2, scm_eval_args (x, env)),
+			      scm_cons2 (t.arg1,
+					 arg2,
+					 scm_eval_args (x, env, proc)),
 			      SCM_ENV (proc));
 	x = SCM_CODE (proc);
 	goto cdrxbegin;
 #endif /* DEVAL */
+      case scm_tcs_cons_gloc:
+	if (SCM_I_OPERATORP (proc))
+	  {
+	    SCM p = (SCM_I_ENTITYP (proc)
+		     ? SCM_ENTITY_PROC_3 (proc)
+		     : SCM_OPERATOR_PROC_3 (proc));
+	    if (SCM_NIMP (p))
+	      {
+		if (SCM_TYP7 (p) == scm_tc7_lsubr_2)
+#ifdef DEVAL
+		  RETURN (SCM_SUBRF (p) (proc, t.arg1,
+					 scm_cons (arg2, SCM_CDDR (debug.info->a.args))))
+#else
+		  RETURN (SCM_SUBRF (p) (proc, t.arg1,
+					 scm_cons (arg2,
+						   scm_eval_args (x, env, proc))))
+#endif
+		else if (SCM_CLOSUREP (p))
+		  {
+#ifdef DEVAL
+		    SCM_SET_ARGSREADY (debug);
+		    debug.info->a.args = scm_cons (proc, debug.info->a.args);
+		    debug.info->a.proc = p;
+		    env = EXTEND_ENV (SCM_CAR (SCM_CODE (p)),
+				      scm_cons2 (proc, t.arg1,
+						 scm_cons (arg2,
+							   SCM_CDDDR (debug.info->a.args))),
+				      SCM_ENV (p));
+#else
+		    env = EXTEND_ENV (SCM_CAR (SCM_CODE (p)),
+				      scm_cons2 (proc, t.arg1,
+						 scm_cons (arg2,
+							   scm_eval_args (x, env, proc))),
+				      SCM_ENV (p));
+#endif
+		    x = SCM_CODE (p);
+		    goto cdrxbegin;
+		  }
+	      }
+	    /* Fall through. */
+	  }
       case scm_tc7_subr_2:
       case scm_tc7_subr_1o:
       case scm_tc7_subr_2o:
@@ -2366,19 +2639,16 @@ evapply:
   }
 #ifdef DEVAL
 exit:
-  if (CHECK_EXIT)
+  if (CHECK_EXIT && SCM_TRAPS_P)
     if (SCM_EXIT_FRAME_P || (SCM_TRACE_P && SCM_TRACED_FRAME_P (debug)))
       {
-	SCM_EXIT_FRAME_P = 0;
-	SCM_TRACE_P = 0;
-	SCM_RESET_DEBUG_MODE;
 	SCM_CLEAR_TRACED_FRAME (debug);
 	if (SCM_CHEAPTRAPS_P)
 	  t.arg1 = scm_make_debugobj (&debug);
 	else
 	  {
 	    scm_make_cont (&t.arg1);
-	    if (safe_setjmp (SCM_JMPBUF (t.arg1)))
+	    if (setjmp (SCM_JMPBUF (t.arg1)))
 	      {
 		proc = SCM_THROW_VALUE (t.arg1);
 		goto ret;
@@ -2565,17 +2835,15 @@ SCM_APPLY (proc, arg1, args)
 #endif
     }
 #ifdef DEVAL
-  if (SCM_ENTER_FRAME_P)
+  if (SCM_ENTER_FRAME_P && SCM_TRAPS_P)
     {
       SCM tmp;
-      SCM_ENTER_FRAME_P = 0;
-      SCM_RESET_DEBUG_MODE;
       if (SCM_CHEAPTRAPS_P)
 	tmp = scm_make_debugobj (&debug);
       else
 	{
 	  scm_make_cont (&tmp);
-	  if (safe_setjmp (SCM_JMPBUF (tmp)))
+	  if (setjmp (SCM_JMPBUF (tmp)))
 	    goto entap;
 	}
       scm_ithrow (scm_i_enter_frame, scm_cons (tmp, SCM_EOL), 0);
@@ -2592,7 +2860,8 @@ tail:
       args = SCM_NULLP (args) ? SCM_UNDEFINED : SCM_CAR (args);
       RETURN (SCM_SUBRF (proc) (arg1, args))
     case scm_tc7_subr_2:
-      SCM_ASRTGO (SCM_NULLP (SCM_CDR (args)), wrongnumargs);
+      SCM_ASRTGO (SCM_NNULLP (args) && SCM_NULLP (SCM_CDR (args)),
+		  wrongnumargs);
       args = SCM_CAR (args);
       RETURN (SCM_SUBRF (proc) (arg1, args))
     case scm_tc7_subr_0:
@@ -2675,11 +2944,28 @@ tail:
 #else
       arg1 = (SCM_UNBNDP (arg1) ? SCM_EOL : scm_cons (arg1, args));
 #endif
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
       if (scm_badargsp (SCM_CAR (SCM_CODE (proc)), arg1))
 	goto wrongnumargs;
 #endif
-      args = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), arg1, SCM_ENV (proc));
+      
+      /* Copy argument list */
+      if (SCM_IMP (arg1))
+	args = arg1;
+      else
+	{
+	  SCM tl = args = scm_cons (SCM_CAR (arg1), SCM_UNSPECIFIED);
+	  while (SCM_NIMP (arg1 = SCM_CDR (arg1))
+		 && SCM_CONSP (arg1))
+	    {
+	      SCM_SETCDR (tl, scm_cons (SCM_CAR (arg1),
+					SCM_UNSPECIFIED));
+	      tl = SCM_CDR (tl);
+	    }
+	  SCM_SETCDR (tl, arg1);
+	}
+      
+      args = EXTEND_ENV (SCM_CAR (SCM_CODE (proc)), args, SCM_ENV (proc));
       proc = SCM_CODE (proc);
       while (SCM_NNULLP (proc = SCM_CDR (proc)))
 	arg1 = EVALCAR (proc, args);
@@ -2702,6 +2988,36 @@ tail:
 #endif
       goto tail;
 #endif
+    case scm_tcs_cons_gloc:
+      if (SCM_I_OPERATORP (proc))
+	{
+#ifdef DEVAL
+	  args = (SCM_UNBNDP(arg1) ? SCM_EOL : debug.vect[0].a.args);
+#else
+	  args = (SCM_UNBNDP(arg1) ? SCM_EOL : scm_cons (arg1, args));
+#endif
+	  arg1 = proc;
+	  proc = (SCM_NULLP (args)
+		  ? (SCM_I_ENTITYP (proc)
+		     ? SCM_ENTITY_PROC_0 (proc)
+		     : SCM_OPERATOR_PROC_0 (proc))
+		  : SCM_NULLP (SCM_CDR (args))
+		  ? (SCM_I_ENTITYP (proc)
+		     ? SCM_ENTITY_PROC_1 (proc)
+		     : SCM_OPERATOR_PROC_1 (proc))
+		  : SCM_NULLP (SCM_CDDR (args))
+		  ? (SCM_I_ENTITYP (proc)
+		     ? SCM_ENTITY_PROC_2 (proc)
+		     : SCM_OPERATOR_PROC_2 (proc))
+		  : (SCM_I_ENTITYP (proc)
+		     ? SCM_ENTITY_PROC_3 (proc)
+		     : SCM_OPERATOR_PROC_3 (proc)));
+#ifdef DEVAL
+	  debug.vect[0].a.proc = proc;
+	  debug.vect[0].a.args = scm_cons (arg1, args);
+#endif
+	  goto tail;
+	}
     wrongnumargs:
       scm_wrong_num_args (proc);
     default:
@@ -2711,18 +3027,16 @@ tail:
     }
 #ifdef DEVAL
 exit:
-  if (CHECK_EXIT)
+  if (CHECK_EXIT && SCM_TRAPS_P)
     if (SCM_EXIT_FRAME_P || (SCM_TRACE_P && SCM_TRACED_FRAME_P (debug)))
       {
-	SCM_EXIT_FRAME_P = 0;
-	SCM_RESET_DEBUG_MODE;
 	SCM_CLEAR_TRACED_FRAME (debug);
 	if (SCM_CHEAPTRAPS_P)
 	  arg1 = scm_make_debugobj (&debug);
 	else
 	  {
 	    scm_make_cont (&arg1);
-	    if (safe_setjmp (SCM_JMPBUF (arg1)))
+	    if (setjmp (SCM_JMPBUF (arg1)))
 	      {
 		proc = SCM_THROW_VALUE (arg1);
 		goto ret;
@@ -2771,7 +3085,7 @@ scm_map (proc, arg1, args)
     }
   args = scm_vector (scm_cons (arg1, args));
   ve = SCM_VELTS (args);
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
   for (i = SCM_LENGTH (args) - 1; i >= 0; i--)
     SCM_ASSERT (SCM_NIMP (ve[i]) && SCM_CONSP (ve[i]), args, SCM_ARG2, s_map);
 #endif
@@ -2816,7 +3130,7 @@ scm_for_each (proc, arg1, args)
     }
   args = scm_vector (scm_cons (arg1, args));
   ve = SCM_VELTS (args);
-#ifndef RECKLESS
+#ifndef SCM_RECKLESS
   for (i = SCM_LENGTH (args) - 1; i >= 0; i--)
     SCM_ASSERT (SCM_NIMP (ve[i]) && SCM_CONSP (ve[i]), args, SCM_ARG2, s_for_each);
 #endif
@@ -2857,8 +3171,10 @@ scm_makprom (code)
 {
   register SCM z;
   SCM_NEWCELL (z);
+  SCM_ENTER_A_SECTION;
   SCM_SETCDR (z, code);
   SCM_SETCAR (z, scm_tc16_promise);
+  SCM_EXIT_A_SECTION;
   return z;
 }
 
@@ -2873,11 +3189,11 @@ prinprom (exp, port, pstate)
      scm_print_state *pstate;
 {
   int writingp = SCM_WRITINGP (pstate);
-  scm_gen_puts (scm_regular_string, "#<promise ", port);
+  scm_puts ("#<promise ", port);
   SCM_SET_WRITINGP (pstate, 1);
   scm_iprin1 (SCM_CDR (exp), port, pstate);
   SCM_SET_WRITINGP (pstate, writingp);
-  scm_gen_putc ('>', port);
+  scm_putc ('>', port);
   return !0;
 }
 
@@ -2924,37 +3240,77 @@ scm_makmmacro (code)
 }
 
 
+SCM_PROC (s_macro_p, "macro?", 1, 0, 0, scm_macro_p);
 
-static int  prinmacro SCM_P ((SCM exp, SCM port, scm_print_state *pstate));
-
-static int 
-prinmacro (exp, port, pstate)
-     SCM exp;
-     SCM port;
-     scm_print_state *pstate;
+SCM
+scm_macro_p (obj)
+     SCM obj;
 {
-  int writingp = SCM_WRITINGP (pstate);
-  if (SCM_CAR (exp) & (3L << 16))
-    scm_gen_puts (scm_regular_string, "#<macro", port);
-  else
-    scm_gen_puts (scm_regular_string, "#<syntax", port);
-  if (SCM_CAR (exp) & (2L << 16))
-    scm_gen_putc ('!', port);
-  scm_gen_putc (' ', port);
-  SCM_SET_WRITINGP (pstate, 1);
-  scm_iprin1 (SCM_CDR (exp), port, pstate);
-  SCM_SET_WRITINGP (pstate, writingp);
-  scm_gen_putc ('>', port);
-  return !0;
+  return (SCM_NIMP (obj) && SCM_TYP16 (obj) == scm_tc16_macro
+	  ? SCM_BOOL_T
+	  : SCM_BOOL_F);
 }
 
+
+SCM_SYMBOL (scm_sym_syntax, "syntax");
+SCM_SYMBOL (scm_sym_macro, "macro");
+SCM_SYMBOL (scm_sym_mmacro, "macro!");
+
+SCM_PROC (s_macro_type, "macro-type", 1, 0, 0, scm_macro_type);
+
+SCM
+scm_macro_type (m)
+     SCM m;
+{
+  if (!(SCM_NIMP (m) && SCM_TYP16 (m) == scm_tc16_macro))
+    return SCM_BOOL_F;
+  switch ((int) (SCM_CAR (m) >> 16))
+    {
+    case 0: return scm_sym_syntax;
+    case 1: return scm_sym_macro;
+    case 2: return scm_sym_mmacro;
+    default: scm_wrong_type_arg (s_macro_type, 1, m);
+    }
+}
+
+
+SCM_PROC (s_macro_name, "macro-name", 1, 0, 0, scm_macro_name);
+
+SCM
+scm_macro_name (m)
+     SCM m;
+{
+  SCM_ASSERT (SCM_NIMP (m) && SCM_TYP16 (m) == scm_tc16_macro,
+	      m,
+	      SCM_ARG1,
+	      s_macro_name);
+  return scm_procedure_name (SCM_CDR (m));
+}
+
+
+SCM_PROC (s_macro_transformer, "macro-transformer", 1, 0, 0, scm_macro_transformer);
+
+SCM
+scm_macro_transformer (m)
+     SCM m;
+{
+  SCM_ASSERT (SCM_NIMP (m) && SCM_TYP16 (m) == scm_tc16_macro,
+	      m,
+	      SCM_ARG1,
+	      s_macro_transformer);
+  return SCM_CLOSUREP (SCM_CDR (m)) ? SCM_CDR (m) : SCM_BOOL_F;
+}
+
+
+
 SCM_PROC(s_force, "force", 1, 0, 0, scm_force);
 
 SCM 
 scm_force (x)
      SCM x;
 {
-  SCM_ASSERT ((SCM_TYP16 (x) == scm_tc16_promise), x, SCM_ARG1, s_force);
+  SCM_ASSERT (SCM_NIMP(x) && SCM_TYP16 (x) == scm_tc16_promise,
+	      x, SCM_ARG1, s_force);
   if (!((1L << 16) & SCM_CAR (x)))
     {
       SCM ans = scm_apply (SCM_CDR (x), SCM_EOL, SCM_EOL);
@@ -2992,7 +3348,7 @@ scm_copy_tree (obj)
   if (SCM_VECTORP (obj))
     {
       scm_sizet i = SCM_LENGTH (obj);
-      ans = scm_make_vector (SCM_MAKINUM (i), SCM_UNSPECIFIED, SCM_UNDEFINED);
+      ans = scm_make_vector (SCM_MAKINUM (i), SCM_UNSPECIFIED);
       while (i--)
 	SCM_VELTS (ans)[i] = scm_copy_tree (SCM_VELTS (obj)[i]);
       return ans;
@@ -3068,17 +3424,6 @@ scm_eval_x (obj)
 	       scm_top_level_env (SCM_CDR (scm_top_level_lookup_closure_var)));
 }
 
-SCM_PROC (s_macro_eval_x, "macro-eval!", 2, 0, 0, scm_macro_eval_x);
-
-SCM
-scm_macro_eval_x (exp, env)
-     SCM exp;
-     SCM env;
-{
-  return scm_eval_3 (exp, 0, env);
-}
-
-
 SCM_PROC (s_definedp, "defined?", 1, 0, 0, scm_definedp);
 
 SCM 
@@ -3099,12 +3444,9 @@ scm_definedp (sym)
       SCM_BOOL_F : SCM_BOOL_T;
 }
 
-static scm_smobfuns promsmob =
-{scm_markcdr, scm_free0, prinprom};
-
-static scm_smobfuns macrosmob =
-{scm_markcdr, scm_free0, prinmacro};
+static scm_smobfuns promsmob = {scm_markcdr, scm_free0, prinprom};
 
+static scm_smobfuns macrosmob = {scm_markcdr, scm_free0};
 
 SCM 
 scm_make_synt (name, macroizer, fcn)
@@ -3138,6 +3480,13 @@ scm_make_synt (name, macroizer, fcn)
 void 
 scm_init_eval ()
 {
+  scm_init_opts (scm_evaluator_traps,
+		 scm_evaluator_trap_table,
+		 SCM_N_EVALUATOR_TRAPS);
+  scm_init_opts (scm_eval_options_interface,
+		 scm_eval_opts,
+		 SCM_N_EVAL_OPTIONS);
+  
   scm_tc16_promise = scm_newsmob (&promsmob);
   scm_tc16_macro = scm_newsmob (&macrosmob);
   scm_i_apply = scm_make_subr ("apply", scm_tc7_lsubr_2, scm_apply);
@@ -3184,6 +3533,8 @@ scm_init_eval ()
 #endif
 
 #include "eval.x"
+
+  scm_add_feature ("delay");
 }
 
 #endif /* !DEVAL */
