@@ -2,7 +2,7 @@
 
 #ifndef SMOBH
 #define SMOBH
-/*	Copyright (C) 1995,1996 Free Software Foundation, Inc.
+/*	Copyright (C) 1995,1996,1998 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -44,6 +44,7 @@
  * If you do not wish that, delete this exception notice.  */
 
 #include "libguile/__scm.h"
+#include "libguile/print.h"
 
 
 typedef struct scm_smobfuns
@@ -58,7 +59,7 @@ typedef struct scm_smobfuns
 
 #define SCM_SMOBNUM(x) (0x0ff & (SCM_CAR(x)>>8))
 
-extern scm_sizet scm_numsmob;
+extern int scm_numsmob;
 extern scm_smobfuns *scm_smobs;
 
 
